#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from pyscf.lib import logger
from pyscf.dft.numint import _dot_ao_dm

def _grid_ao2mo (mol, ao, mo_coeff, non0tab=None, shls_slice=None,
        ao_loc=None):
    '''ao[deriv,grid,AO].mo_coeff[AO,MO]->mo[deriv,grid,MO]
    ASSUMES that ao is in data layout (deriv,AO,grid) in row-major order!
    mo is returned in data layout (deriv,MO,grid) in row-major order '''
    nderiv, ngrid, _ = ao.shape
    nmo = mo_coeff.shape[-1]
    mo = np.empty ((nderiv,nmo,ngrid), dtype=mo_coeff.dtype, order='C')
    mo = mo.transpose (0,2,1)
    if shls_slice is None: shls_slice = (0, mol.nbas)
    if ao_loc is None: ao_loc = mol.ao_loc_nr ()
    for ideriv in range (nderiv):
        ao_i = ao[ideriv,:,:]
        mo[ideriv] = _dot_ao_dm (mol, ao_i, mo_coeff, non0tab, shls_slice,
            ao_loc, out=mo[ideriv])
    return mo


def get_ontop_pair_density (ot, rho, ao, cascm2, mo_cas, deriv=0,
        non0tab=None):
    r'''Compute the on-top pair density and its derivatives on a grid:

    Pi(r) = i(r)*j(r)*k(r)*l(r)*d_ijkl / 2
          = rho[0](r)*rho[1](r) + i(r)*j(r)*k(r)*l(r)*l_ijkl / 2

    Args:
        ot : on-top pair density functional object
        rho : ndarray of shape (2,*,ngrids)
            Contains spin-separated density [and derivatives]. The dm1s
            underlying these densities must correspond to the dm1s/dm1
            in the expression for cascm2 below.
        ao : ndarray of shape (*, ngrids, nao)
            contains values of aos [and derivatives]
        cascm2 : ndarray of shape [ncas,]*4
            contains spin-summed two-body cumulant density matrix in an
            active-orbital basis given by mo_cas:
                cm2[u,v,x,y] = dm2[u,v,x,y] - dm1[u,v]*dm1[x,y]
                               + dm1s[0][u,y]*dm1s[0][x,v]
                               + dm1s[1][u,y]*dm1s[1][x,v]
            where dm1 = dm1s[0] + dm1s[1]. The cumulant (cm2) has no
            nonzero elements for any index outside the active space,
            unlike the density matrix (dm2), which formally has elements
            involving uncorrelated, doubly-occupied ``core'' orbitals
            which are not usually computed explicitly:
                dm2[i,i,u,v] = dm2[u,v,i,i] = 2*dm1[u,v]
                dm2[u,i,i,v] = dm2[i,v,u,i] = -dm1[u,v]
        mo_cas : ndarray of shape (nao, ncas)
            molecular-orbital coefficients for active-space orbitals

    Kwargs:
        deriv : derivative order through which to calculate.
            deriv > 1 not implemented
        non0tab : as in pyscf.dft.gen_grid and pyscf.dft.numint

    Returns : ndarray of shape (*,ngrids)
        The on-top pair density and its derivatives if requested
        deriv = 0 : value (1d array)
        deriv = 1 : value, d/dx, d/dy, d/dz
        deriv = 2 : value, d/dx, d/dy, d/dz, d^2/d|r1-r2|^2_(r1=r2)
    '''
    # Fix dimensionality of rho and ao
    rho_reshape = False
    ao_reshape = False
    if rho.ndim == 2:
        rho_reshape = True
        rho = rho.reshape(rho.shape[0], 1, rho.shape[1])
    if ao.ndim == 2:
        ao_reshape = True
        ao = ao.reshape(1, ao.shape[0], ao.shape[1])

    # First cumulant and derivatives (chain rule! product rule!)
    t0 = (logger.process_clock (), logger.perf_counter ())
    Pi_shape = ((1,4,5)[deriv], rho.shape[-1])
    Pi = np.zeros(Pi_shape, dtype=rho.dtype)
    Pi[0] = rho[0,0] * rho[1,0]
    if deriv > 0:
        assert (rho.shape[1] >= 4), rho.shape
        assert (ao.shape[0] >= 4), ao.shape
        for ideriv in range(1,4):
            Pi[ideriv] = rho[0,ideriv]*rho[1,0] + rho[0,0]*rho[1,ideriv]
    if deriv > 1:
        assert (rho.shape[1] >= 6), rho.shape
        assert (ao.shape[0] >= 10), ao.shape
        Pi[4] = -(rho[:,1:4].sum (0).conjugate () * rho[:,1:4].sum (0)).sum (0)
        Pi[4] /= 4.0
        Pi[4] += rho[0,0]*(rho[1,4]/4 + rho[0,5]*2)
        Pi[4] += rho[1,0]*(rho[0,4]/4 + rho[1,5]*2)
    t0 = logger.timer_debug1 (ot, 'otpd first cumulant', *t0)

    # Second cumulant and derivatives (chain rule! product rule!)
    # dot, tensordot, and sum are hugely faster than np.einsum
    # but whether or when they actually multithread is unclear
    # Update 05/11/2020: ao is actually stored in row-major order
    # = (deriv,AOs,grids).
    grid2amo = _grid_ao2mo (ot.mol, ao, mo_cas, non0tab=non0tab)
    t0 = logger.timer (ot, 'otpd ao2mo', *t0)
    gridkern = np.zeros (grid2amo.shape + (grid2amo.shape[2],),
        dtype=grid2amo.dtype)
    gridkern[0] = grid2amo[0,:,:,np.newaxis] * grid2amo[0,:,np.newaxis,:]
    # r_0ai,  r_0aj  -> r_0aij
    wrk0 = np.tensordot (gridkern[0], cascm2, axes=2)
    # r_0aij, P_ijkl -> P_0akl
    Pi[0] += (gridkern[0] * wrk0).sum ((1,2)) / 2
    # r_0aij, P_0aij -> P_0a
    t0 = logger.timer_debug1 (ot, 'otpd second cumulant 0th derivative', *t0)
    if deriv > 0:
        for ideriv in range (1, 4):
            # Fourfold tensor symmetry ijkl = klij = jilk = lkji
            # & product rule -> factor of 4
            gridkern[ideriv] = (grid2amo[ideriv,:,:,np.newaxis]
                * grid2amo[0,:,np.newaxis,:])
            # r_1ai,  r_0aj  -> r_1aij
            Pi[ideriv] += (gridkern[ideriv] * wrk0).sum ((1,2)) * 2
            # r_1aij, P_0aij -> P_1a
            t0 = logger.timer_debug1 (ot, 'otpd second cumulant 1st derivative'
                ' ({})'.format (ideriv), *t0)
    if deriv > 1: # The fifth slot is allocated to the "off-top Laplacian,"
        # i.e., nabla_(r1-r2)^2 Pi(r1,r2)|(r1=r2)
        # nabla_off^2 Pi = 1/2 d^ik_jl * ([nabla_r^2 phi_i] phi_j phi_k phi_l
        # + {1 - p_jk - p_jl}[nabla_r phi_i . nabla_r phi_j] phi_k phi_l)
        # using four-fold symmetry a lot! be careful!
        XX, YY, ZZ = 4, 7, 9
        gridkern[4]  = (grid2amo[[XX,YY,ZZ],:,:,np.newaxis].sum (0)
            * grid2amo[0,:,np.newaxis,:])
        # r_2ai, r_0aj -> r_2aij
        gridkern[4] += (grid2amo[1:4,:,:,np.newaxis]
            * grid2amo[1:4,:,np.newaxis,:]).sum (0)
        # r_1ai, r_1aj -> r_2aij
        wrk1 = np.tensordot (gridkern[1:4], cascm2, axes=2)
        # r_1aij, P_ijkl -> P_1akl
        Pi[4] += (gridkern[4] * wrk0).sum ((1,2)) / 2
        # r_2aij, P_0aij -> P_2a
        Pi[4] -= ((gridkern[1:4] + gridkern[1:4].transpose (0, 1, 3, 2))
            * wrk1).sum ((0,2,3)) / 2
        # r_1aij, P_1aij -> P_2a
        t0 = logger.timer (ot, 'otpd second cumulant off-top Laplacian', *t0)

    # Unfix dimensionality of rho, ao, and Pi
    # if Pi.shape[0] == 1:
    if rho_reshape:
        Pi = Pi.reshape (Pi.shape[1])
        rho = rho.reshape (rho.shape[0], rho.shape[2])
    if ao_reshape:
        ao = ao.reshape (ao.shape[1], ao.shape[2])

    return Pi

def density_orbital_derivative (ot, ncore, ncas, casdm1s, cascm2, rho, mo,
        deriv=0, non0tab=None):
    '''Compute the half-transformed density and 3/4-transformed pair-
    density matrix:

    D_i(r) = sum_j D_ij phi_j(r)
    d_i(r) = sum_jkl d_ijkl phi_j(r) phi_k(r) phi_l(r)

    so that, for instance, the derivative with respect to orbital
    rotations of the density and pair density are

    d/dk_ij rho(r) = phi_i(r) D_j(r) - phi_j(r) D_k(r)
    d/dk_ij Pi(r) = i(r) d_j(r) - j(r) d_i(r)

    and the derivatives with respect to nuclear displacement are

    drho/dRA|(r not in A) = -sum_(mu in A) phi'_mu(r) D_mu(r)
    drho/dRA|(r in A) = +sum_(mu not in A) phi'_mu(r) D_mu(r)
    dPi/dRA|(r not in A) = -sum_(mu in A) phi'_mu(r) d_mu(r)
    dPi/dRA|(r in A) = +sum_(mu not in A) phi'_mu(r) d_mu(r)

    There is a mismatch between the ndarray shape and the data layout in
    the arg `mo` and the two return arrays. For performance reasons, the
    grid index should be contiguous in memory for every array with a
    a grid dimension. However, by convention, in PySCF ndarrays with
    grid and orbital indices place the latter index in the last
    position, the former in the second-to-last position, and any other
    before that in row-major order. That is, for a four-index array of
    this type, transpose (2,3,1,0) gives a contiguous column-major array,
    and transpose (0,1,3,2) results in a contiguous row-major array.

    Args:
        ot : object of :class:`otfnal`
            The on-top density functional containing grid information
        ncore : integer
            Number of doubly-occupied core orbitals
        ncas : integer
            Number of active orbitals
        casdm1s : ndarray of shape (2,ncas,ncas)
            Spin-separated one-body reduced density matrix
        cascm2 : ndarray of shape [ncas,]*4
            Spin-summed cumulant of two-body reduced density matrix
        rho : ndarray of shape (2,*,ngrids)
            Spin-separated density (and derivatives) on a grid
        mo : ndarray of shape (*,ngrids,nmo)
            Molecular orbitals (and derivatives) evaluated on a grid.
            Data stride must be 0>2>1; i.e., mo.transpose (0,2,1) must
            be a contiguous row-major array.

    Kwargs:
        deriv : integer
            Order of derivatives of half-transformed density to compute;
            i.e., 0 for LDA and 1 for GGA
        non0tab : 2D boolean array
            Mask array to determine whether densities are considered to
            be numerically zero

    Returns:
        drho : ndarray of shape (2,*,ngrids,nmo)
            Half-transformed density matrix on a grid and its
            derivatives. Data stride is 0>1>3>2; i.e.,
            drho.transpose (0,1,3,2) is a contiguous row-major array.
        dPi : ndarray of shape (*,ngrids,nmo)
            3/4-transformed pair density matrix on a grid and its
            derivatives. Data stride is 0>2>1; i.e.,
            dPi.transpose (0,2,1) is a contiguous row-major array.
    '''
    nocc = ncore + ncas
    nderiv_Pi = (1,4)[int (deriv)]
    nmo = mo.shape[-1]
    ngrids = rho.shape[-1]

    # Fix dimensionality of rho and mo
    if rho.ndim == 2:
        rho = rho.reshape (rho.shape[0], 1, rho.shape[1])
    if mo.ndim == 2:
        mo = mo.reshape (1, mo.shape[0], mo.shape[1])

    # First cumulant and derivatives
    dm1s_mo = np.stack ([np.eye (nmo, dtype=casdm1s[0].dtype),]*2, axis=0)
    dm1s_mo[:,ncore:nocc,ncore:nocc] = casdm1s
    dm1s_mo[:,nocc:,:] = 0
    dm1s_mo[:,:,nocc:] = 0

    drho = np.stack ([_grid_ao2mo (ot.mol, mo, dm1, non0tab=non0tab)
        for dm1 in dm1s_mo], axis=0).transpose (0,1,3,2)
    dPi = np.zeros ((nderiv_Pi, nmo, ngrids), dtype=rho.dtype)
    dPi[0] = ((drho[0][0] * rho[1,0,None,:])
           + (rho[0,0,None,:] * drho[1][0]))
    if deriv > 0:
        for ideriv in range(1,4):
            dPi[ideriv] = ((drho[0][ideriv]*rho[1,0,None,:])
                         + (drho[0][0]*rho[1,ideriv,None,:])
                         + (rho[0,ideriv,None,:]*drho[1][0])
                         + (rho[0,0,None,:]*drho[1][ideriv]))
    if deriv > 1:
        raise NotImplementedError ("Colle-Salvetti type orbital+grid "
            "derivatives")

    # Second cumulant and derivatives
    mo_cas = mo[:,:,ncore:nocc]
    gridkern = np.zeros (mo_cas.shape + (mo_cas.shape[2],), dtype=mo_cas.dtype)
    gridkern[0] = mo_cas[0,:,:,np.newaxis] * mo_cas[0,:,np.newaxis,:]
    # r_0ai,  r_0aj  -> r_0aij
    wrk0 = np.tensordot (gridkern[0], cascm2, axes=2)
    # r_0aij, P_ijkl -> P_0akl
    dPi[0,ncore:nocc] += (mo_cas[0][:,None,:] * wrk0).sum (2).T
    # r_0aj,  P_0aij -> P_0ai
    if deriv > 0:
        for ideriv in range (1, 4):
            dPi[ideriv,ncore:nocc] += (mo_cas[ideriv][:,None,:]
                * wrk0).sum (2).T
            # r_1aj,  P_0aij -> P_1ai
            gridkern[ideriv] = (mo_cas[ideriv,:,:,np.newaxis]
                * mo_cas[0,:,np.newaxis,:])
            gridkern[ideriv] += gridkern[ideriv].transpose (0,2,1)
            # r_1ai,  r_0aj  -> r_1aij
        for ideriv in range (1, 4):
            wrk0 = np.tensordot (gridkern[ideriv], cascm2, axes=2)
            # r_1aij, P_ijkl -> P_1akl
            dPi[ideriv,ncore:nocc] += (mo_cas[0][:,None,:] * wrk0).sum (2).T
            # r_0aj,  P_1aij -> P_1ai
    if deriv > 1:
        raise NotImplementedError ("Colle-Salvetti type orbital+grid "
            "derivatives")

    # Unfix dimensionality of rho, ao, and Pi
    dPi = dPi.transpose (0,2,1)
    drho = drho.transpose (0,1,3,2)

    return drho, dPi





# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import incore
from . import outcore
from . import fft
from . import aft
from . import df
from . import mdf
from .df import DF, GDF
from .rsdf import RSDF, RSGDF
from .mdf import MDF
from .aft import AFTDF
from .fft import FFTDF
from pyscf.df.addons import aug_etb
from .incore import make_auxcell

# For backward compatibility
pwdf = aft
PWDF = AFTDF



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Non-relativistic restricted Hartree-Fock with point group symmetry.

The symmetry are not handled in a separate data structure.  Note that during
the SCF iteration,  the orbitals are grouped in terms of symmetry irreps.
But the orbitals in the result are sorted based on the orbital energies.
Function symm.label_orb_symm can be used to detect the symmetry of the
molecular orbitals.
'''

from functools import reduce
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import symm
from pyscf.lib import logger
from pyscf.scf import hf
from pyscf.scf import uhf
from pyscf.scf import rohf
from pyscf.scf import chkfile
from pyscf.lib.exceptions import PointGroupSymmetryError
from pyscf import __config__

WITH_META_LOWDIN = getattr(__config__, 'scf_analyze_with_meta_lowdin', True)
MO_BASE = getattr(__config__, 'MO_BASE', 1)


# mo_energy, mo_coeff, mo_occ are all in nosymm representation

def analyze(mf, verbose=logger.DEBUG, with_meta_lowdin=WITH_META_LOWDIN,
            **kwargs):
    '''Analyze the given SCF object:  print orbital energies, occupancies;
    print orbital coefficients; Occupancy for each irreps; Mulliken population analysis
    '''
    from pyscf.lo import orth
    from pyscf.tools import dump_mat
    mol = mf.mol
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    mo_coeff = mf.mo_coeff
    ovlp_ao = mf.get_ovlp()
    log = logger.new_logger(mf, verbose)
    if log.verbose >= logger.NOTE:
        mf.dump_scf_summary(log)

        nirrep = len(mol.irrep_id)
        orbsym = mf.get_orbsym(mo_coeff, ovlp_ao)
        wfnsym = 0
        noccs = [sum(orbsym[mo_occ>0]==ir) for ir in mol.irrep_id]
        if mol.groupname in symm.param.POINTGROUP:
            log.note('Wave-function symmetry = %s',
                     symm.irrep_id2name(mol.groupname, wfnsym))
        else:
            # TODO: check wave function symmetry for ('SO3', 'Dooh', 'Coov') etc
            log.note('Wave-function symmetry id = %s', wfnsym)
        log.note('occupancy for each irrep:  ' + (' %4s'*nirrep), *mol.irrep_name)
        log.note('                           ' + (' %4d'*nirrep), *noccs)
        log.note('**** MO energy ****')
        irname_full = {}
        for k,ir in enumerate(mol.irrep_id):
            irname_full[ir] = mol.irrep_name[k]
        irorbcnt = {}
        for k, j in enumerate(orbsym):
            if j in irorbcnt:
                irorbcnt[j] += 1
            else:
                irorbcnt[j] = 1
            log.note('MO #%d (%s #%d), energy= %.15g occ= %g',
                     k+MO_BASE, irname_full[j], irorbcnt[j],
                     mo_energy[k], mo_occ[k])

    if log.verbose >= logger.DEBUG:
        label = mol.ao_labels()
        molabel = []
        irorbcnt = {}
        for k, j in enumerate(orbsym):
            if j in irorbcnt:
                irorbcnt[j] += 1
            else:
                irorbcnt[j] = 1
            molabel.append('#%-d(%s #%d)' %
                           (k+MO_BASE, irname_full[j], irorbcnt[j]))
        if with_meta_lowdin:
            log.debug(' ** MO coefficients (expansion on meta-Lowdin AOs) **')
            orth_coeff = orth.orth_ao(mol, 'meta_lowdin', s=ovlp_ao)
            c = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mo_coeff))
        else:
            log.debug(' ** MO coefficients (expansion on AOs) **')
            c = mo_coeff
        dump_mat.dump_rec(mf.stdout, c, label, molabel, start=MO_BASE, **kwargs)

    dm = mf.make_rdm1(mo_coeff, mo_occ)
    if with_meta_lowdin:
        pop_and_charge = mf.mulliken_meta(mol, dm, s=ovlp_ao, verbose=log)
    else:
        pop_and_charge = mf.mulliken_pop(mol, dm, s=ovlp_ao, verbose=log)
    dip = mf.dip_moment(mol, dm, verbose=log)
    return pop_and_charge, dip

def get_irrep_nelec(mol, mo_coeff, mo_occ, s=None):
    '''Electron numbers for each irreducible representation.

    Args:
        mol : an instance of :class:`Mole`
            To provide irrep_id, and spin-adapted basis
        mo_coeff : 2D ndarray
            Regular orbital coefficients, without grouping for irreps
        mo_occ : 1D ndarray
            Regular occupancy, without grouping for irreps

    Returns:
        irrep_nelec : dict
            The number of electrons for each irrep {'ir_name':int,...}.

    Examples:

    >>> mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='ccpvdz', symmetry=True, verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.scf()
    -76.016789472074251
    >>> scf.hf_symm.get_irrep_nelec(mol, mf.mo_coeff, mf.mo_occ)
    {'A1': 6, 'A2': 0, 'B1': 2, 'B2': 2}
    '''
    orbsym = get_orbsym(mol, mo_coeff, s, False)
    irrep_nelec = {mol.irrep_name[k]: int(sum(mo_occ[orbsym==ir]))
                        for k, ir in enumerate(mol.irrep_id)}
    return irrep_nelec

def canonicalize(mf, mo_coeff, mo_occ, fock=None):
    '''Canonicalization diagonalizes the Fock matrix in occupied, open,
    virtual subspaces separatedly (without change occupancy).
    '''
    mol = mf.mol
    if not mol.symmetry:
        raise RuntimeError('mol.symmetry not enabled')
    if fock is None:
        dm = mf.make_rdm1(mo_coeff, mo_occ)
        fock = mf.get_hcore() + mf.get_veff(mf.mol, dm)
    coreidx = mo_occ == 2
    viridx = mo_occ == 0
    openidx = ~(coreidx | viridx)
    mo = numpy.empty_like(mo_coeff)
    mo_e = numpy.empty(mo_occ.size)

    if getattr(mo_coeff, 'orbsym', None) is not None:
        orbsym = mo_coeff.orbsym
        irreps = set(orbsym)
        for ir in irreps:
            idx0 = orbsym == ir
            for idx1 in (coreidx, openidx, viridx):
                idx = idx0 & idx1
                if numpy.count_nonzero(idx) > 0:
                    orb = mo_coeff[:,idx]
                    f1 = reduce(numpy.dot, (orb.conj().T, fock, orb))
                    e, c = scipy.linalg.eigh(f1)
                    mo[:,idx] = numpy.dot(mo_coeff[:,idx], c)
                    mo_e[idx] = e
    else:
        s = mf.get_ovlp()
        for idx in (coreidx, openidx, viridx):
            if numpy.count_nonzero(idx) > 0:
                orb = mo_coeff[:,idx]
                f1 = reduce(numpy.dot, (orb.conj().T, fock, orb))
                e, c = scipy.linalg.eigh(f1)
                c = numpy.dot(mo_coeff[:,idx], c)
                mo[:,idx] = _symmetrize_canonicalization_(mf, e, c, s)
                mo_e[idx] = e
        orbsym = mf.get_orbsym(mo, s)

    mo = lib.tag_array(mo, orbsym=orbsym)
    return mo_e, mo

def _symmetrize_canonicalization_(mf, mo_energy, mo_coeff, s):
    '''Restore symmetry for canonicalized orbitals
    '''
    def search_for_degeneracy(mo_energy):
        idx = numpy.where(abs(mo_energy[1:] - mo_energy[:-1]) < 1e-6)[0]
        return numpy.unique(numpy.hstack((idx, idx+1)))

    mol = mf.mol
    degidx = search_for_degeneracy(mo_energy)
    logger.debug1(mf, 'degidx %s', degidx)
    if degidx.size > 0:
        esub = mo_energy[degidx]
        csub = mo_coeff[:,degidx]
        scsub = numpy.dot(s, csub)
        emin = abs(esub).min() * .5
        es = []
        cs = []
        for i,ir in enumerate(mol.irrep_id):
            so = mol.symm_orb[i]
            sosc = numpy.dot(so.conj().T, scsub)
            s_ir = reduce(numpy.dot, (so.conj().T, s, so))
            fock_ir = numpy.dot(sosc*esub, sosc.conj().T)
            mo_energy, u = mf._eigh(fock_ir, s_ir)
            idx = abs(mo_energy) > emin
            es.append(mo_energy[idx])
            cs.append(numpy.dot(mol.symm_orb[i], u[:,idx]))
        es = numpy.hstack(es).round(7)
        idx = numpy.argsort(es, kind='stable')
        assert (numpy.allclose(es[idx], esub.round(7)))
        mo_coeff[:,degidx] = numpy.hstack(cs)[:,idx]
    return mo_coeff

def so2ao_mo_coeff(so, irrep_mo_coeff):
    '''Transfer the basis of MO coefficients, from symmetry-adapted basis to AO basis
    '''
    return numpy.hstack([numpy.dot(so[ir],irrep_mo_coeff[ir])
                         for ir in range(so.__len__())])

def check_irrep_nelec(mol, irrep_nelec, nelec):
    for irname in irrep_nelec:
        if irname not in mol.irrep_name:
            msg = 'Molecule does not have irrep %s defined in irrep_nelec.' % irname
            raise ValueError(msg)

    float_irname = []
    fix_na = 0
    fix_nb = 0
    free_irrep_norbs = []
    for i, irname in enumerate(mol.irrep_name):
        if irname in irrep_nelec:
            if isinstance(irrep_nelec[irname], (int, numpy.integer)):
                nelecb = irrep_nelec[irname] // 2
                neleca = irrep_nelec[irname] - nelecb
            else:
                neleca, nelecb = irrep_nelec[irname]
            if not (isinstance(neleca, (int, numpy.integer)) and
                    isinstance(nelecb, (int, numpy.integer))):
                raise ValueError('irrep_nelec value must be integer')
            norb = mol.symm_orb[i].shape[1]
            if neleca > norb or nelecb > norb:
                msg =('More electrons than orbitals for irrep %s '
                      'nelec = %d + %d, norb = %d' %
                      (irname, neleca,nelecb, norb))
                raise ValueError(msg)
            fix_na += neleca
            fix_nb += nelecb
        else:
            float_irname.append(irname)
            free_irrep_norbs.append(mol.symm_orb[i].shape[1])

    if isinstance(nelec, (int, numpy.integer)):
        nelecb = nelec // 2
        neleca = nelec - nelecb
    else:
        neleca, nelecb = nelec
    fix_ne = fix_na + fix_nb
    float_neleca = neleca - fix_na
    float_nelecb = nelecb - fix_nb
    free_norb = sum(free_irrep_norbs)
    if fix_na > neleca or fix_nb > nelecb:
        msg =('More electrons defined by irrep_nelec than total num electrons.\n'
              f'total num electrons = ({neleca}, {nelecb})\n'
              f'num electrons defined by irrep_nelec = ({fix_na}, {fix_nb})\n'
              f'irrep_nelec = {irrep_nelec}')
        raise ValueError(msg)
    else:
        logger.info(mol, 'Freeze %d electrons in irreps %s',
                    fix_ne, list(irrep_nelec.keys()))

    if len(set(float_irname)) == 0 and fix_ne != mol.nelectron:
        msg =('Num electrons defined by irrep_nelec != total num electrons. '
              'mol.nelectron = %d  irrep_nelec = %s' %
              (mol.nelectron, irrep_nelec))
        raise ValueError(msg)
    elif float_neleca > free_norb or float_nelecb > free_norb:
        raise ValueError('Not enough orbitals for (%d, %d) electrons in irreps %s '
                         '(irrep_norb: %s)' %
                         (float_neleca, float_nelecb, ' '.join(float_irname), free_norb))
    else:
        logger.info(mol, '    %d free electrons in irreps %s',
                    mol.nelectron-fix_ne, ' '.join(float_irname))
    return fix_na, fix_nb, float_irname

#TODO: force Dooh, Doov orbitals E1gx/E1gy ... using the same coefficients
#TODO: force SO3 orbitals p+0,p-1,p+1, ... using the same coefficients
def eig(mf, h, s, symm_orb=None, irrep_id=None):
    '''Solve generalized eigenvalue problem, for each irrep.  The
    eigenvalues and eigenvectors are not sorted to ascending order.
    Instead, they are grouped based on irreps.
    '''
    mol = mf.mol
    if symm_orb is None or irrep_id is None:
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')
        symm_orb = mol.symm_orb
        irrep_id = mol.irrep_id

    nirrep = symm_orb.__len__()
    h = symm.symmetrize_matrix(h, symm_orb)
    s = symm.symmetrize_matrix(s, symm_orb)
    cs = []
    es = []
    orbsym = []
    if getattr(mol, 'groupname', None) in ('Dooh', 'Coov'):
        for ir in range(nirrep):
            irrep_1d = irrep_id[ir] in (0, 1, 4, 5)
            irrep_2dx = irrep_id[ir] % 2 == 0
            if irrep_1d or irrep_2dx:
                e, c = mf._eigh(h[ir], s[ir])
                cs.append(c)
                es.append(e)
                orbsym.append([irrep_id[ir]] * e.size)

            if not irrep_1d and irrep_2dx:
                # force 2D irreps using the same coefficients
                irrep_conj = irrep_id[ir] ^ 1
                assert irrep_id[ir+1] == irrep_conj
                cs.append(c)
                es.append(e)
                orbsym.append([irrep_conj] * e.size)
    else:
        for ir in range(nirrep):
            e, c = mf._eigh(h[ir], s[ir])
            cs.append(c)
            es.append(e)
            orbsym.append([irrep_id[ir]] * e.size)
    e = numpy.hstack(es)
    c = so2ao_mo_coeff(symm_orb, cs)
    c = lib.tag_array(c, orbsym=numpy.hstack(orbsym))
    return e, c

def get_orbsym(mol, mo_coeff, s=None, check=False, symm_orb=None, irrep_id=None):
    if getattr(mo_coeff, 'orbsym', None) is not None:
        return mo_coeff.orbsym

    if symm_orb is None or irrep_id is None:
        symm_orb = mol.symm_orb
        irrep_id = mol.irrep_id
    if mo_coeff is None:
        orbsym = numpy.hstack([[ir] * symm_orb[i].shape[1]
                               for i, ir in enumerate(irrep_id)])
    else:
        orbsym = symm.label_orb_symm(mol, irrep_id, symm_orb,
                                     mo_coeff, s, check)
    return orbsym

def map_degeneracy(mo_energy, orbsym):
    '''Find degeneracy correspondence for cylindrical symmetry'''
    norb = orbsym.size
    ex_mask = numpy.isin(orbsym % 10, (0, 2, 5, 7))

    degen_mapping = numpy.arange(norb)
    for ir_ex in set(orbsym[ex_mask]):
        if ir_ex in (0, 1, 5, 4):
            continue

        if ir_ex % 2 == 0:
            ir_ey = ir_ex + 1
        else:
            ir_ey = ir_ex - 1
        ex_idx = numpy.where(orbsym == ir_ex)[0]
        ey_idx = numpy.where(orbsym == ir_ey)[0]
        if ex_idx.size != ey_idx.size:
            raise PointGroupSymmetryError('Degenerated orbitals required')
        mo_ex = mo_energy[ex_idx].round(7)
        mo_ey = mo_energy[ey_idx].round(7)
        mapping = numpy.where(abs(mo_ex[:,None] - mo_ey) < 1e-6)[1]
        if mapping.size != ex_idx.size:
            raise PointGroupSymmetryError('Degenerated orbitals required')
        degen_mapping[ex_idx] = ey_idx[mapping]
        degen_mapping[ey_idx] = ex_idx[mapping.argsort()]

    return degen_mapping

def get_wfnsym(mf, mo_coeff=None, mo_occ=None, orbsym=None):
    if mo_occ is None:
        mo_occ = mf.mo_occ
    if orbsym is None:
        orbsym = mf.get_orbsym(mo_coeff)

    if mf.mol.groupname == 'SO3':
        if numpy.any(orbsym > 7):
            logger.warn(mf, 'Wave-function symmetry for %s not supported. '
                        'Wfn symmetry is mapped to D2h/C2v group.',
                        mf.mol.groupname)

    wfnsym = 0
    for ir in orbsym[mo_occ == 1] % 10:
        wfnsym ^= ir

    if mf.mol.groupname in ('Dooh', 'Coov'):
        orb_l = (orbsym // 10) * 2
        e1_mask = numpy.isin(orbsym % 10, (2, 3, 6, 7))
        orb_l[e1_mask] += 1
        ey_mask = numpy.isin(orbsym % 10, (1, 3, 4, 6))
        orb_l[ey_mask] *= -1
        a_l = orb_l[mo_occ != 0]
        b_l = orb_l[mo_occ == 2]
        wfn_momentum = a_l.sum() + b_l.sum()
        wfnsym += (abs(wfn_momentum) // 2) * 10

    return wfnsym


class SymAdaptedRHF(hf.RHF):
    __doc__ = hf.SCF.__doc__ + '''
    Attributes for symmetry allowed RHF:
        irrep_nelec : dict
            Specify the number of electrons for particular irrep {'ir_name':int,...}.
            For the irreps not listed in this dict, the program will choose the
            occupancy based on the orbital energies.

    Examples:

    >>> mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='ccpvdz', symmetry=True, verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.scf()
    -76.016789472074251
    >>> mf.get_irrep_nelec()
    {'A1': 6, 'A2': 0, 'B1': 2, 'B2': 2}
    >>> mf.irrep_nelec = {'A2': 2}
    >>> mf.scf()
    -72.768201804695622
    >>> mf.get_irrep_nelec()
    {'A1': 6, 'A2': 2, 'B1': 2, 'B2': 0}
    '''

    _keys = {'irrep_nelec'}

    def __init__(self, mol):
        hf.RHF.__init__(self, mol)
        # number of electrons for each irreps
        self.irrep_nelec = {} # {'ir_name':int,...}

    def build(self, mol=None):
        if mol is None: mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')
        check_irrep_nelec(mol, self.irrep_nelec, self.mol.nelectron)
        return hf.RHF.build(self, mol)

    eig = eig

    def get_grad(self, mo_coeff, mo_occ, fock=None):
        g = hf.RHF.get_grad(self, mo_coeff, mo_occ, fock)
        occidx = mo_occ > 0
        viridx = ~occidx
        orbsym = self.get_orbsym(mo_coeff)
        sym_forbid = orbsym[viridx].reshape(-1,1) != orbsym[occidx]
        g[sym_forbid.ravel()] = 0
        return g

    def get_occ(self, mo_energy=None, mo_coeff=None):
        ''' We assumed mo_energy are grouped by symmetry irreps, (see function
        self.eig). The orbitals are sorted after SCF.
        '''
        if mo_energy is None: mo_energy = self.mo_energy
        mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')

        orbsym = self.get_orbsym(mo_coeff)
        mo_occ = numpy.zeros_like(mo_energy)
        rest_idx = numpy.ones(mo_occ.size, dtype=bool)
        nelec_fix = 0
        for i, ir in enumerate(mol.irrep_id):
            irname = mol.irrep_name[i]
            if irname in self.irrep_nelec:
                ir_idx = numpy.where(orbsym == ir)[0]
                n = self.irrep_nelec[irname]
                occ_sort = numpy.argsort(mo_energy[ir_idx].round(9), kind='stable')
                occ_idx  = ir_idx[occ_sort[:n//2]]
                mo_occ[occ_idx] = 2
                nelec_fix += n
                rest_idx[ir_idx] = False
        nelec_float = mol.nelectron - nelec_fix
        assert (nelec_float >= 0)
        if nelec_float > 0:
            rest_idx = numpy.where(rest_idx)[0]
            occ_sort = numpy.argsort(mo_energy[rest_idx].round(9), kind='stable')
            occ_idx  = rest_idx[occ_sort[:nelec_float//2]]
            mo_occ[occ_idx] = 2

        vir_idx = (mo_occ==0)
        if self.verbose >= logger.INFO and numpy.count_nonzero(vir_idx) > 0:
            ehomo = max(mo_energy[~vir_idx])
            elumo = min(mo_energy[ vir_idx])
            noccs = []
            for i, ir in enumerate(mol.irrep_id):
                irname = mol.irrep_name[i]
                ir_idx = (orbsym == ir)

                noccs.append(int(mo_occ[ir_idx].sum()))
                if ehomo in mo_energy[ir_idx]:
                    irhomo = irname
                if elumo in mo_energy[ir_idx]:
                    irlumo = irname
            logger.info(self, 'HOMO (%s) = %.15g  LUMO (%s) = %.15g',
                        irhomo, ehomo, irlumo, elumo)

            logger.debug(self, 'irrep_nelec = %s', noccs)
            _dump_mo_energy(mol, mo_energy, mo_occ, ehomo, elumo, orbsym,
                            verbose=self.verbose)
        return mo_occ

    def _finalize(self):
        hf.RHF._finalize(self)

        # sort MOs wrt orbital energies, it should be done last.
        # Using mergesort because it is stable. We don't want to change the
        # ordering of the symmetry labels when two orbitals are degenerated.
        o_sort = numpy.argsort(self.mo_energy[self.mo_occ> 0].round(9), kind='stable')
        v_sort = numpy.argsort(self.mo_energy[self.mo_occ==0].round(9), kind='stable')
        idx = numpy.arange(self.mo_energy.size)
        idx = numpy.hstack((idx[self.mo_occ> 0][o_sort],
                            idx[self.mo_occ==0][v_sort]))
        self.mo_energy = self.mo_energy[idx]
        self.mo_occ = self.mo_occ[idx]
        orbsym = self.get_orbsym(self.mo_coeff)
        orbsym = orbsym[idx]
        degen_mapping = None
        if self.mol.groupname in ('Dooh', 'Coov'):
            try:
                degen_mapping = map_degeneracy(self.mo_energy, orbsym)
            except PointGroupSymmetryError:
                logger.warn(self, 'Orbital degeneracy broken')
        if degen_mapping is None:
            self.mo_coeff = lib.tag_array(self.mo_coeff[:,idx], orbsym=orbsym)
        else:
            self.mo_coeff = lib.tag_array(
                self.mo_coeff[:,idx], orbsym=orbsym, degen_mapping=degen_mapping)
        if self.chkfile:
            chkfile.dump_scf(self.mol, self.chkfile, self.e_tot, self.mo_energy,
                             self.mo_coeff, self.mo_occ, overwrite_mol=False)
        return self

    @lib.with_doc(analyze.__doc__)
    def analyze(self, verbose=None, with_meta_lowdin=WITH_META_LOWDIN,
                **kwargs):
        if verbose is None: verbose = self.verbose
        return analyze(self, verbose, with_meta_lowdin, **kwargs)

    @lib.with_doc(get_irrep_nelec.__doc__)
    def get_irrep_nelec(self, mol=None, mo_coeff=None, mo_occ=None, s=None):
        if mol is None: mol = self.mol
        if mo_occ is None: mo_occ = self.mo_occ
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if s is None: s = self.get_ovlp()
        return get_irrep_nelec(mol, mo_coeff, mo_occ, s)

    def get_orbsym(self, mo_coeff=None, s=None):
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if getattr(mo_coeff, 'orbsym', None) is not None:
            return mo_coeff.orbsym
        if s is None:
            s = self.get_ovlp()
        return numpy.asarray(get_orbsym(self.mol, mo_coeff, s))
    orbsym = property(get_orbsym)

    get_wfnsym = get_wfnsym
    wfnsym = property(get_wfnsym)

    canonicalize = canonicalize

    to_gpu = lib.to_gpu

RHF = SymAdaptedRHF


class SymAdaptedROHF(rohf.ROHF):
    __doc__ = hf.SCF.__doc__ + '''
    Attributes for symmetry allowed ROHF:
        irrep_nelec : dict
            Specify the number of alpha/beta electrons for particular irrep
            {'ir_name':(int,int), ...}.
            For the irreps not listed in these dicts, the program will choose the
            occupancy based on the orbital energies.

    Examples:

    >>> mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='ccpvdz', symmetry=True, charge=1, spin=1, verbose=0)
    >>> mf = scf.RHF(mol)
    >>> mf.scf()
    -75.619358861084052
    >>> mf.get_irrep_nelec()
    {'A1': (3, 3), 'A2': (0, 0), 'B1': (1, 1), 'B2': (1, 0)}
    >>> mf.irrep_nelec = {'B1': (1, 0)}
    >>> mf.scf()
    -75.425669486776457
    >>> mf.get_irrep_nelec()
    {'A1': (3, 3), 'A2': (0, 0), 'B1': (1, 0), 'B2': (1, 1)}
    '''

    _keys = {'irrep_nelec'}

    def __init__(self, mol):
        rohf.ROHF.__init__(self, mol)
        self.irrep_nelec = {}
# use _irrep_doccs and _irrep_soccs help self.eig to compute orbital energy,
# do not overwrite them
        self._irrep_doccs = []
        self._irrep_soccs = []

    def dump_flags(self, verbose=None):
        rohf.ROHF.dump_flags(self, verbose)
        if self.irrep_nelec:
            logger.info(self, 'irrep_nelec %s', self.irrep_nelec)
        return self

    def build(self, mol=None):
        if mol is None: mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')
        fix_na, fix_nb = check_irrep_nelec(mol, self.irrep_nelec, self.nelec)[:2]
        alpha_open = beta_open = False
        for ne in self.irrep_nelec.values():
            if not isinstance(ne, (int, numpy.integer)):
                alpha_open |= ne[0] > ne[1]
                beta_open  |= ne[0] < ne[1]

        frozen_spin = fix_na - fix_nb
        if ((alpha_open and beta_open) or
            (0 < mol.spin < frozen_spin) or (frozen_spin < 0 < mol.spin) or
            (frozen_spin < mol.spin < 0) or (mol.spin < 0 < frozen_spin)):
            raise ValueError('Low-spin configuration was found in '
                             'the irrep_nelec input. ROHF does not '
                             'support low-spin configuration.')
        return hf.RHF.build(self, mol)

    @lib.with_doc(eig.__doc__)
    def eig(self, fock, s):
        e, c = eig(self, fock, s)
        if getattr(fock, 'focka', None) is not None:
            mo_ea = numpy.einsum('pi,pi->i', c, fock.focka.dot(c))
            mo_eb = numpy.einsum('pi,pi->i', c, fock.fockb.dot(c))
            e = lib.tag_array(e, mo_ea=mo_ea, mo_eb=mo_eb)
        return e, c

    def get_grad(self, mo_coeff, mo_occ, fock=None):
        g = rohf.ROHF.get_grad(self, mo_coeff, mo_occ, fock)
        occidxa = mo_occ > 0
        occidxb = mo_occ == 2
        viridxa = ~occidxa
        viridxb = ~occidxb
        uniq_var_a = viridxa.reshape(-1,1) & occidxa
        uniq_var_b = viridxb.reshape(-1,1) & occidxb

        orbsym = self.get_orbsym(mo_coeff)
        sym_forbid = orbsym.reshape(-1,1) != orbsym
        sym_forbid = sym_forbid[uniq_var_a | uniq_var_b]
        g[sym_forbid.ravel()] = 0
        return g

    def get_occ(self, mo_energy=None, mo_coeff=None):
        if mo_energy is None: mo_energy = self.mo_energy
        mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')

        if getattr(mo_energy, 'mo_ea', None) is not None:
            mo_ea = mo_energy.mo_ea
            mo_eb = mo_energy.mo_eb
        else:
            mo_ea = mo_eb = mo_energy
        nmo = mo_ea.size
        mo_occ = numpy.zeros(nmo)
        orbsym = self.get_orbsym(mo_coeff)

        rest_idx = numpy.ones(mo_occ.size, dtype=bool)
        neleca_fix = 0
        nelecb_fix = 0
        for i, ir in enumerate(mol.irrep_id):
            irname = mol.irrep_name[i]
            if irname in self.irrep_nelec:
                ir_idx = numpy.where(orbsym == ir)[0]
                if isinstance(self.irrep_nelec[irname], (int, numpy.integer)):
                    nelecb = self.irrep_nelec[irname] // 2
                    neleca = self.irrep_nelec[irname] - nelecb
                else:
                    neleca, nelecb = self.irrep_nelec[irname]
                if neleca > nelecb:
                    ncore, nopen = nelecb, neleca - nelecb
                else:
                    ncore, nopen = neleca, nelecb - neleca
                mo_occ[ir_idx] = rohf._fill_rohf_occ(mo_energy[ir_idx],
                                                     mo_ea[ir_idx], mo_eb[ir_idx],
                                                     ncore, nopen)
                neleca_fix += neleca
                nelecb_fix += nelecb
                rest_idx[ir_idx] = False

        nelec_float = mol.nelectron - neleca_fix - nelecb_fix
        assert (nelec_float >= 0)
        if len(rest_idx) > 0:
            rest_idx = numpy.where(rest_idx)[0]
            nopen = abs(mol.spin - (neleca_fix - nelecb_fix))
            ncore = (nelec_float - nopen)//2
            mo_occ[rest_idx] = rohf._fill_rohf_occ(mo_energy[rest_idx],
                                                   mo_ea[rest_idx], mo_eb[rest_idx],
                                                   ncore, nopen)

        nocc, ncore = self.nelec
        nopen = nocc - ncore
        vir_idx = (mo_occ==0)
        if self.verbose >= logger.INFO and nocc < nmo and ncore > 0:
            ehomo = max(mo_energy[~vir_idx])
            elumo = min(mo_energy[ vir_idx])
            ndoccs = []
            nsoccs = []
            for i, ir in enumerate(mol.irrep_id):
                irname = mol.irrep_name[i]
                ir_idx = (orbsym == ir)

                ndoccs.append(numpy.count_nonzero(mo_occ[ir_idx]==2))
                nsoccs.append(numpy.count_nonzero(mo_occ[ir_idx]==1))
                if ehomo in mo_energy[ir_idx]:
                    irhomo = irname
                if elumo in mo_energy[ir_idx]:
                    irlumo = irname

            # to help self.eigh compute orbital energy
            self._irrep_doccs = ndoccs
            self._irrep_soccs = nsoccs

            logger.info(self, 'HOMO (%s) = %.15g  LUMO (%s) = %.15g',
                        irhomo, ehomo, irlumo, elumo)

            logger.debug(self, 'double occ irrep_nelec = %s', ndoccs)
            logger.debug(self, 'single occ irrep_nelec = %s', nsoccs)
            #_dump_mo_energy(mol, mo_energy, mo_occ, ehomo, elumo, orbsym,
            #                verbose=self.verbose)
            if nopen > 0:
                core_idx = mo_occ == 2
                open_idx = mo_occ == 1
                vir_idx = mo_occ == 0
                logger.debug(self, '                  Roothaan           | alpha              | beta')
                logger.debug(self, '  Highest 2-occ = %18.15g | %18.15g | %18.15g',
                             max(mo_energy[core_idx]),
                             max(mo_ea[core_idx]), max(mo_eb[core_idx]))
                logger.debug(self, '  Lowest 0-occ =  %18.15g | %18.15g | %18.15g',
                             min(mo_energy[vir_idx]),
                             min(mo_ea[vir_idx]), min(mo_eb[vir_idx]))
                for i in numpy.where(open_idx)[0]:
                    logger.debug(self, '  1-occ =         %18.15g | %18.15g | %18.15g',
                                 mo_energy[i], mo_ea[i], mo_eb[i])

            numpy.set_printoptions(threshold=nmo)
            logger.debug(self, '  Roothaan mo_energy =\n%s', mo_energy)
            logger.debug1(self, '  alpha mo_energy =\n%s', mo_ea)
            logger.debug1(self, '  beta  mo_energy =\n%s', mo_eb)
            numpy.set_printoptions(threshold=1000)
        return mo_occ

    def make_rdm1(self, mo_coeff=None, mo_occ=None, **kwargs):
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if mo_occ is None:
            mo_occ = self.mo_occ
        mo_a = mo_coeff[:,mo_occ>0]
        mo_b = mo_coeff[:,mo_occ==2]
        dm_a = numpy.dot(mo_a, mo_a.conj().T)
        dm_b = numpy.dot(mo_b, mo_b.conj().T)
        return lib.tag_array((dm_a, dm_b), mo_coeff=mo_coeff, mo_occ=mo_occ)

    def _finalize(self):
        rohf.ROHF._finalize(self)

        # sort MOs wrt orbital energies, it should be done last.
        # Using mergesort because it is stable. We don't want to change the
        # ordering of the symmetry labels when two orbitals are degenerated.
        c_sort = numpy.argsort(self.mo_energy[self.mo_occ==2].round(9), kind='stable')
        o_sort = numpy.argsort(self.mo_energy[self.mo_occ==1].round(9), kind='stable')
        v_sort = numpy.argsort(self.mo_energy[self.mo_occ==0].round(9), kind='stable')
        idx = numpy.arange(self.mo_energy.size)
        idx = numpy.hstack((idx[self.mo_occ==2][c_sort],
                            idx[self.mo_occ==1][o_sort],
                            idx[self.mo_occ==0][v_sort]))
        if getattr(self.mo_energy, 'mo_ea', None) is not None:
            mo_ea = self.mo_energy.mo_ea[idx]
            mo_eb = self.mo_energy.mo_eb[idx]
            self.mo_energy = lib.tag_array(self.mo_energy[idx],
                                           mo_ea=mo_ea, mo_eb=mo_eb)
        else:
            self.mo_energy = self.mo_energy[idx]
        self.mo_occ = self.mo_occ[idx]
        orbsym = self.get_orbsym(self.mo_coeff)
        orbsym = orbsym[idx]
        degen_mapping = None
        if self.mol.groupname in ('Dooh', 'Coov'):
            try:
                degen_mapping = map_degeneracy(self.mo_energy, orbsym)
            except PointGroupSymmetryError:
                logger.warn(self, 'Orbital degeneracy broken')
        if degen_mapping is None:
            self.mo_coeff = lib.tag_array(self.mo_coeff[:,idx], orbsym=orbsym)
        else:
            self.mo_coeff = lib.tag_array(
                self.mo_coeff[:,idx], orbsym=orbsym, degen_mapping=degen_mapping)
        if self.chkfile:
            chkfile.dump_scf(self.mol, self.chkfile, self.e_tot, self.mo_energy,
                             self.mo_coeff, self.mo_occ, overwrite_mol=False)
        return self

    def analyze(self, verbose=None, with_meta_lowdin=WITH_META_LOWDIN,
                **kwargs):
        if verbose is None: verbose = self.verbose
        from pyscf.lo import orth
        from pyscf.tools import dump_mat
        mol = self.mol
        mo_energy = self.mo_energy
        mo_occ = self.mo_occ
        mo_coeff = self.mo_coeff
        ovlp_ao = self.get_ovlp()
        log = logger.new_logger(self, verbose)
        if log.verbose >= logger.NOTE:
            self.dump_scf_summary(log)

            nirrep = len(mol.irrep_id)
            orbsym = self.get_orbsym(mo_coeff)
            irreps = numpy.asarray(mol.irrep_id)
            ndoccs = numpy.count_nonzero(irreps[:,None] == orbsym[mo_occ==2], axis=1)
            nsoccs = numpy.count_nonzero(irreps[:,None] == orbsym[mo_occ==1], axis=1)

            wfnsym = 0
            # wfn symmetry is determined by the odd number of electrons in each irrep
            for k in numpy.where(nsoccs % 2 == 1)[0]:
                ir_in_d2h = mol.irrep_id[k] % 10  # convert to D2h irreps
                wfnsym ^= ir_in_d2h

            if mol.groupname in ('SO3', 'Dooh', 'Coov'):
                # TODO: check wave function symmetry
                log.note('Wave-function symmetry = %s', mol.groupname)
            else:
                log.note('Wave-function symmetry = %s',
                         symm.irrep_id2name(mol.groupname, wfnsym))
            log.note('occupancy for each irrep:  ' + (' %4s'*nirrep),
                     *mol.irrep_name)
            log.note('double occ                 ' + (' %4d'*nirrep), *ndoccs)
            log.note('single occ                 ' + (' %4d'*nirrep), *nsoccs)
            log.note('**** MO energy ****')
            irname_full = {}
            for k,ir in enumerate(mol.irrep_id):
                irname_full[ir] = mol.irrep_name[k]
            irorbcnt = {}
            if getattr(mo_energy, 'mo_ea', None) is not None:
                mo_ea = mo_energy.mo_ea
                mo_eb = mo_energy.mo_eb
                log.note('                          Roothaan           | alpha              | beta')
                for k, j in enumerate(orbsym):
                    if j in irorbcnt:
                        irorbcnt[j] += 1
                    else:
                        irorbcnt[j] = 1
                    log.note('MO #%-4d(%-3s #%-2d) energy= %-18.15g | %-18.15g | %-18.15g occ= %g',
                             k+MO_BASE, irname_full[j], irorbcnt[j],
                             mo_energy[k], mo_ea[k], mo_eb[k], mo_occ[k])
            else:
                for k, j in enumerate(orbsym):
                    if j in irorbcnt:
                        irorbcnt[j] += 1
                    else:
                        irorbcnt[j] = 1
                    log.note('MO #%-3d (%s #%-2d), energy= %-18.15g occ= %g',
                             k+MO_BASE, irname_full[j], irorbcnt[j],
                             mo_energy[k], mo_occ[k])

        if log.verbose >= logger.DEBUG:
            label = mol.ao_labels()
            molabel = []
            irorbcnt = {}
            for k, j in enumerate(orbsym):
                if j in irorbcnt:
                    irorbcnt[j] += 1
                else:
                    irorbcnt[j] = 1
                molabel.append('#%-d(%s #%d)' %
                               (k+MO_BASE, irname_full[j], irorbcnt[j]))
            if with_meta_lowdin:
                log.debug(' ** MO coefficients (expansion on meta-Lowdin AOs) **')
                orth_coeff = orth.orth_ao(mol, 'meta_lowdin', s=ovlp_ao)
                c = reduce(numpy.dot, (orth_coeff.conj().T, ovlp_ao, mo_coeff))
            else:
                log.debug(' ** MO coefficients (expansion on AOs) **')
                c = mo_coeff
            dump_mat.dump_rec(self.stdout, c, label, molabel, start=MO_BASE, **kwargs)

        dm = self.make_rdm1(mo_coeff, mo_occ)
        if with_meta_lowdin:
            pop_and_charge = self.mulliken_meta(mol, dm, s=ovlp_ao, verbose=log)
        else:
            pop_and_charge = self.mulliken_pop(mol, dm, s=ovlp_ao, verbose=log)
        dip = self.dip_moment(mol, dm, verbose=log)
        return pop_and_charge, dip

    def get_irrep_nelec(self, mol=None, mo_coeff=None, mo_occ=None, s=None):
        from pyscf.scf import uhf_symm
        if mol is None: mol = self.mol
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if mo_occ is None: mo_occ = self.mo_occ
        if isinstance(mo_coeff, numpy.ndarray) and mo_coeff.ndim == 2:
            mo_coeff = (mo_coeff, mo_coeff)
        if isinstance(mo_occ, numpy.ndarray) and mo_occ.ndim == 1:
            mo_occ = (numpy.array(mo_occ>0, dtype=numpy.double),
                      numpy.array(mo_occ==2, dtype=numpy.double))
        if s is None:
            s = self.get_ovlp()
        return uhf_symm.get_irrep_nelec(mol, mo_coeff, mo_occ, s)

    @lib.with_doc(canonicalize.__doc__)
    def canonicalize(self, mo_coeff, mo_occ, fock=None):
        if getattr(fock, 'focka', None) is None:
            fock = self.get_fock(dm=self.make_rdm1(mo_coeff, mo_occ))
        mo_e, mo_coeff = canonicalize(self, mo_coeff, mo_occ, fock)
        mo_ea = numpy.einsum('pi,pi->i', mo_coeff, fock.focka.dot(mo_coeff))
        mo_eb = numpy.einsum('pi,pi->i', mo_coeff, fock.fockb.dot(mo_coeff))
        mo_e = lib.tag_array(mo_e, mo_ea=mo_ea, mo_eb=mo_eb)
        return mo_e, mo_coeff

    get_orbsym = SymAdaptedRHF.get_orbsym
    orbsym = property(get_orbsym)

    get_wfnsym = get_wfnsym
    wfnsym = property(get_wfnsym)

    to_gpu = lib.to_gpu

ROHF = SymAdaptedROHF


def _dump_mo_energy(mol, mo_energy, mo_occ, ehomo, elumo, orbsym, title='',
                    verbose=logger.DEBUG):
    log = logger.new_logger(mol, verbose)
    for i, ir in enumerate(mol.irrep_id):
        irname = mol.irrep_name[i]
        ir_idx = (orbsym == ir)
        nso = numpy.count_nonzero(ir_idx)
        nocc = numpy.count_nonzero(mo_occ[ir_idx])
        e_ir = mo_energy[ir_idx]
        if nocc == 0:
            log.debug('%s%s nocc = 0', title, irname)
        elif nocc == nso:
            log.debug('%s%s nocc = %d  HOMO = %.15g',
                      title, irname, nocc, e_ir[nocc-1])
        else:
            log.debug('%s%s nocc = %d  HOMO = %.15g  LUMO = %.15g',
                      title, irname, nocc, e_ir[nocc-1], e_ir[nocc])
            if e_ir[nocc-1]+1e-3 > elumo:
                log.warn('%s%s HOMO %.15g > system LUMO %.15g',
                         title, irname, e_ir[nocc-1], elumo)
            if e_ir[nocc] < ehomo+1e-3:
                log.warn('%s%s LUMO %.15g < system HOMO %.15g',
                         title, irname, e_ir[nocc], ehomo)
        log.debug('   mo_energy = %s', e_ir)


class HF1e(ROHF):
    scf = hf._hf1e_scf


del (WITH_META_LOWDIN)



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Matthew Hennefarth <matthew.hennefarth@gmail.com>,
#          Aniruddha Seal <aniruddhaseal2011@gmail.com>

import os
import numpy as np

from pyscf import data
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad.rhf import GradientsBase


class Frame:
    '''Basic class to hold information at each time step of a MD simulation

    Attributes:
        ekin : float
            Kinetic energy
        epot : float
            Potential energy (electronic energy)
        etot : float
            Total energy, sum of the potential and kinetic energy
        coord : 2D array with shape (natm, 3)
            Geometry of the system at the current time step
        veloc : 2D array with shape (natm, 3)
            Velocities of the system at the current time step
        time : float
            Time for which this frame represents
    '''

    def __init__(self,
                 ekin=None,
                 epot=None,
                 coord=None,
                 veloc=None,
                 time=None):
        self.ekin = ekin
        self.epot = epot
        self.etot = self.ekin + self.epot
        self.coord = coord
        self.veloc = veloc
        self.time = time


def _toframe(integrator):
    '''Convert an _Integrator to a Frame given current saved data.

    Args:
        integrator : md.integrator._Integrator object

    Returns:
        Frame with all data taken from the integrator.
    '''
    return Frame(ekin=integrator.ekin,
                 epot=integrator.epot,
                 coord=integrator.mol.atom_coords(),
                 veloc=integrator.veloc,
                 time=integrator.time)


def _write(dev, mol, vec, atmlst=None):
    '''Format output of molecular vector quantity.

    Args:
        dev : lib.logger.Logger object
        mol : gto.mol object
        vec : 2D array with shape (mol.natm, 3)
        atmlst : array of indices to pull atoms from.
            Must be smaller than mol.natm.
    '''
    if atmlst is None:
        atmlst = range(mol.natm)
    dev.stdout.write('         x                y                z\n')
    for k, ia in enumerate(atmlst):
        dev.stdout.write(
            '%d %s  %15.10f  %15.10f  %15.10f\n' %
            (ia, mol.atom_symbol(ia), vec[k, 0], vec[k, 1], vec[k, 2]))


def kernel(integrator, verbose=logger.NOTE):
    log = logger.new_logger(integrator, verbose)
    t0 = (logger.process_clock(), logger.perf_counter())
    log.debug('Start BOMD')

    integrator.mol = integrator.mol.copy()

    # Begin the iterations to run molecular dynamics
    t1 = t0
    for iteration, frame in enumerate(integrator):
        log.note('----------- %s final geometry -----------',
                 integrator.__class__.__name__)
        _write(integrator, integrator.mol, frame.coord)
        log.note('----------------------------------------------')
        log.note('------------ %s final velocity -----------',
                 integrator.__class__.__name__)
        _write(integrator, integrator.mol, frame.veloc)
        log.note('----------------------------------------------')
        log.note('Ekin = %17.13f', frame.ekin)
        log.note('Epot = %17.13f', frame.epot)
        log.note('Etot = %17.13f', frame.etot)

        t1 = log.timer('BOMD iteration %d' % iteration, *t1)

    t0 = log.timer('BOMD', *t0)
    return integrator


class _Integrator(lib.StreamObject):
    '''Integrator abstract base class. Should never be directly constructed,
    but inherited from.

    Args:
        method : lib.GradScanner, rhf.GradientsBase instance, or
        has nuc_grad_method method.
            Method by which to compute the energy gradients and energies
            in order to propagate the equations of motion. Realistically,
            it can be any callable object such that it returns the energy
            and potential energy gradient when given a mol.

    Attributes:
        incore_anyway : bool
            If true, then it will save every frame in memory.
            False, no frames are saved.

        veloc : ndarray
            Initial velocity for the simulation. Values should be given
            in atomic units (Bohr/a.u.). Dimensions should be (natm, 3) such as

             [[x1, y1, z1],
             [x2, y2, z2],
             [x3, y3, z3]]

        verbose : int
            Print level

        steps : int
            Number of steps to take when the kernel or run function is called.

        dt : float
            Time between steps. Given in atomic units.

        stdout : file object
            Default is self.scanner.mol.stdout.

        data_output : file object
            Stream to write energy and temperature to
            during the course of the simulation.

        trajectory_output : file object
            Stream to write the trajectory to during the course of the
            simulation. Written in xyz format.

        frames : ndarray of Frames or None
            If incore_anyway is true, then this will hold a list of frames
            corresponding to the simulation trajectory.

        epot : float
            Potential energy of the last time step during the simulation.

        ekin : float
            Kinetic energy of the last time step during the simulation

        time : float
            Time of the last step during the simulation.

        callback : function(envs_dict) => None
            Callback function takes one dict as the argument which is
            generated by the builtin function :func:`locals`, so that the
            callback function can access all local variables in the current
            environment.
    '''

    def __init__(self, method, **kwargs):

        if isinstance(method, lib.GradScanner):
            self.scanner = method
        elif isinstance(method, GradientsBase):
            self.scanner = method.as_scanner()
        elif getattr(method, 'nuc_grad_method', None):
            self.scanner = method.nuc_grad_method().as_scanner()
        else:
            raise NotImplementedError('Nuclear gradients of %s not available' %
                                      method)

        self.mol = self.scanner.mol
        self.stdout = self.mol.stdout
        self.incore_anyway = self.mol.incore_anyway
        self.veloc = None
        self.verbose = self.mol.verbose
        self.steps = 1
        self.dt = 10
        self.frames = None
        self.epot = None
        self.ekin = None
        self.time = 0
        self.data_output = None
        self.trajectory_output = None
        self.callback = None

        # Cache the masses into a list, they will be in atomic units
        self._masses = None

        self.__dict__.update(kwargs)

    def kernel(self, veloc=None, steps=None, dump_flags=True, verbose=None):
        '''Runs the molecular dynamics simulation.

        Args:
            veloc : ndarray
                Initial velocity for the simulation. Values should be given
                in atomic units (Bohr/a.u.). Dimensions should be (natm, 3)
                such as

                [[x1, y1, z1],
                 [x2, y2, z2],
                 [x3, y3, z3]]

            steps : int
                Number of steps to take when the kernel or run function
                is called.

            dump_flags : bool
                Print flags to output.

            verbose : int
                Print level

        Returns:
            _Integrator with final epot, ekin, temp,
            mol, and veloc of the simulation.
        '''

        if veloc is not None:
            self.veloc = veloc

        if steps is not None:
            self.steps = steps

        if verbose is None:
            verbose = self.verbose

        # Default velocities are 0 if none specified
        if self.veloc is None:
            self.veloc = np.full((self.mol.natm, 3), 0.0)

        # Store the masses into a cached variable,
        # so we don't have to keep looking them up
        self._masses = np.array([
            data.elements.COMMON_ISOTOPE_MASSES[m] * data.nist.AMU2AU
            for m in self.mol.atom_charges()])

        # avoid opening data_output file twice
        if type(self.data_output) is str:
            if self.verbose > logger.QUIET:
                if os.path.isfile(self.data_output):
                    print('overwrite data output file: %s' %
                          self.data_output)
                else:
                    print('data output file: %s' % self.data_output)

            if self.data_output == '/dev/null':
                self.data_output = open(os.devnull, 'w')

            else:
                self.data_output = open(self.data_output, 'w')
                self.data_output.write(
                    'time          Epot                 Ekin                 '
                    'Etot                 T\n'
                )

        # avoid opening trajectory_output file twice
        if type(self.trajectory_output) is str:
            if self.verbose > logger.QUIET:
                if os.path.isfile(self.trajectory_output):
                    print('overwrite energy output file: %s' %
                          self.trajectory_output)
                else:
                    print('trajectory output file: %s' %
                          self.trajectory_output)

            if self.trajectory_output == '/dev/null':
                self.trajectory_output = open(os.devnull, 'w')

            else:
                self.trajectory_output = open(self.trajectory_output, 'w')

        log = logger.new_logger(self, verbose)
        self.check_sanity()

        if dump_flags and self.verbose > logger.NOTE:
            self.dump_input()

        return kernel(self, verbose=log)

    def dump_input(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('')
        log.info('******** BOMD flags ********')
        log.info('dt = %f', self.dt)
        log.info('Iterations = %d', self.steps)
        log.info('                   Initial Velocity                  ')
        log.info('             vx              vy              vz')
        for i, (e, v) in enumerate(zip(self.mol.elements, self.veloc)):
            log.info('%d %s    %.8E  %.8E  %.8E', i, e, *v)

    def check_sanity(self):
        assert self.time >= 0
        assert self.dt > 0
        assert self.steps > 0
        assert self.veloc is not None
        assert self.veloc.shape == (self.mol.natm, 3)
        assert self.scanner is not None

        return self

    def compute_kinetic_energy(self):
        '''Compute the kinetic energy of the current frame.'''
        # TODO, can make this cleaner by removing an explicit zip and
        # try to leverage numpy vectors
        energy = 0
        for v, m in zip(self.veloc, self._masses):
            energy += 0.5 * m * np.linalg.norm(v) ** 2

        return energy

    def temperature(self):
        '''Returns the temperature of the system'''
        # checked against ORCA for linear and non-linear molecules
        dof = 3 * len(self.mol.atom_coords())

        # Temp = 2/(3*k*N_f) * KE
        #      = 2/(3*k*N_f)*\sum_i (1/2 m_i v_i^2)
        return ((2 * self.ekin) / (
                dof * data.nist.BOLTZMANN / data.nist.HARTREE2J))

    def __iter__(self):
        self._step = 0
        self._log = logger.new_logger(self, self.verbose)
        return self

    def __next__(self):
        if self._step < self.steps:
            if self._log.verbose >= lib.logger.NOTE:
                self._log.note('\nBOMD Time %.2f', self.time)

            current_frame = self._next()

            if self.incore_anyway:
                self.frames.append(current_frame)

            if self.data_output is not None:
                self._write_data()

            if self.trajectory_output is not None:
                self._write_coord()

            if callable(self.callback):
                mol = self.mol
                scanner = self.scanner
                self.callback(locals())

            self._step += 1
            self.time += self.dt

            return current_frame

        else:
            raise StopIteration

    def _next(self):
        '''Determines the next step in the molecular dynamics simulation.
        Integrates to the next time step. Must be implemented in derived
        classes.

        Returns: 'Frame' which contains the new geometry, velocity,
        time step, and energy.
        '''
        raise NotImplementedError('Method Not Implemented')

    def _write_data(self):
        '''Writes out the potential, kinetic, and total energy, temperature to the
        self.data_output stream. '''

        output = '%8.2f  %.12E  %.12E  %.12E %3.4f' % (self.time,
                                                       self.epot,
                                                       self.ekin,
                                                       self.ekin + self.epot,
                                                       self.temperature())

        # We follow OM of writing all the states at the end of the line
        if getattr(self.scanner.base, 'e_states', None) is not None:
            if len(self.scanner.base.e_states) > 1:
                for e in self.scanner.base.e_states:
                    output += '  %.12E' % e

        self.data_output.write(output + '\n')

        # If we don't flush, there is a possibility of losing data
        self.data_output.flush()

    def _write_coord(self):
        '''Writes out the current geometry to the self.trajectory_output
        stream in xyz format. '''
        self.trajectory_output.write('%s\nMD Time %.2f\n' %
                                     (self.mol.natm, self.time))
        self.trajectory_output.write(self.mol.tostring(format='raw') + '\n')

        # If we don't flush, there is a possibility of losing data
        self.trajectory_output.flush()


class VelocityVerlet(_Integrator):
    '''Velocity Verlet algorithm

    Args:
        method : lib.GradScanner or rhf.GradientsBase instance, or
        has nuc_grad_method method.
            Method by which to compute the energy gradients and energies
            in order to propagate the equations of motion. Realistically,
            it can be any callable object such that it returns the energy
            and potential energy gradient when given a mol.

    Attributes:
        accel : ndarray
            Current acceleration for the simulation. Values are given
            in atomic units (Bohr/a.u.^2). Dimensions is (natm, 3) such as

             [[x1, y1, z1],
             [x2, y2, z2],
             [x3, y3, z3]]
    '''

    def __init__(self, method, **kwargs):
        super().__init__(method, **kwargs)
        self.accel = None

    def _next(self):
        '''Computes the next frame of the simulation and sets all internal
         variables to this new frame. First computes the new geometry,
         then the next acceleration, and finally the velocity, all according
         to the Velocity Verlet algorithm.

        Returns:
            The next frame of the simulation.
        '''

        # If no acceleration, compute that first, and then go
        # onto the next step
        if self.accel is None:
            next_epot, next_accel = self._compute_accel()

        else:
            self.mol.set_geom_(self._next_geometry(), unit='B')
            self.mol.build()
            next_epot, next_accel = self._compute_accel()
            self.veloc = self._next_velocity(next_accel)

        self.epot = next_epot
        self.ekin = self.compute_kinetic_energy()
        self.accel = next_accel

        return _toframe(self)

    def _compute_accel(self):
        '''Given the current geometry, computes the acceleration
        for each atom.'''
        e_tot, grad = self.scanner(self.mol)
        if not self.scanner.converged:
            raise RuntimeError('Gradients did not converge!')

        a = -1 * grad / self._masses.reshape(-1, 1)
        return e_tot, a

    def _next_geometry(self):
        '''Computes the next geometry using the Velocity Verlet algorithm. The
        necessary equations of motion for the position is
            r(t_i+1) = r(t_i) + /delta t * v(t_i) + 0.5(/delta t)^2 a(t_i)
        '''
        return self.mol.atom_coords() + self.dt * self.veloc + \
            0.5 * (self.dt ** 2) * self.accel

    def _next_velocity(self, next_accel):
        '''Compute the next velocity using the Velocity Verlet algorithm. The
        necessary equations of motion for the velocity is
            v(t_i+1) = v(t_i) + 0.5(a(t_i+1) + a(t_i))'''
        return self.veloc + 0.5 * self.dt * (self.accel + next_accel)


class NVTBerendson(_Integrator):
    '''Berendsen (constant N, V, T) molecular dynamics

    Args:
        method : lib.GradScanner or rhf.GradientsMixin instance, or
        has nuc_grad_method method.
            Method by which to compute the energy gradients and energies
            in order to propagate the equations of motion. Realistically,
            it can be any callable object such that it returns the energy
            and potential energy gradient when given a mol.

        T      : float
            Target temperature for the NVT Ensemble. Given in K.

        taut   : float
            Time constant for Berendsen temperature coupling.
            Given in atomic units.

    Attributes:
        accel : ndarray
            Current acceleration for the simulation. Values are given
            in atomic units (Bohr/a.u.^2). Dimensions is (natm, 3) such as

             [[x1, y1, z1],
             [x2, y2, z2],
             [x3, y3, z3]]
    '''

    def __init__(self, method, T, taut, **kwargs):
        self.T = T
        self.taut = taut
        self.accel = None
        super().__init__(method, **kwargs)

    def _next(self):
        '''Computes the next frame of the simulation and sets all internal
         variables to this new frame. First computes the new geometry,
         then the next acceleration, and finally the velocity, all according
         to the Velocity Verlet algorithm.

        Returns:
            The next frame of the simulation.
        '''

        # If no acceleration, compute that first, and then go
        # onto the next step
        if self.accel is None:
            next_epot, next_accel = self._compute_accel()

        else:
            self._scale_velocities()
            self.mol.set_geom_(self._next_geometry(), unit='B')
            self.mol.build()
            next_epot, next_accel = self._compute_accel()
            self.veloc = self._next_velocity(next_accel)

        self.epot = next_epot
        self.ekin = self.compute_kinetic_energy()
        self.accel = next_accel

        return _toframe(self)

    def _compute_accel(self):
        '''Given the current geometry, computes the acceleration
        for each atom.'''
        e_tot, grad = self.scanner(self.mol)
        if not self.scanner.converged:
            raise RuntimeError('Gradients did not converge!')

        a = -1 * grad / self._masses.reshape(-1, 1)
        return e_tot, a

    def _scale_velocities(self):
        '''NVT Berendsen velocity scaling
        v_rescale(t) = v(t) * (1 + ((T_target/T - 1)
                            * (/delta t / taut)))^(0.5)
        '''
        tautscl = self.dt / self.taut
        scl_temp = np.sqrt(1.0 + (self.T / self.temperature() - 1.0) * tautscl)

        # Limit the velocity scaling to reasonable values
        # (taken from ase md/nvtberendson.py)
        if scl_temp > 1.1:
            scl_temp = 1.1
        if scl_temp < 0.9:
            scl_temp = 0.9

        self.veloc = self.veloc * scl_temp
        return

    def _next_geometry(self):
        '''Computes the next geometry using the Velocity Verlet algorithm. The
        necessary equations of motion for the position is
            r(t_i+1) = r(t_i) + /delta t * v(t_i) + 0.5(/delta t)^2 a(t_i)
        '''
        return self.mol.atom_coords() + self.dt * self.veloc + \
            0.5 * (self.dt ** 2) * self.accel

    def _next_velocity(self, next_accel):
        '''Compute the next velocity using the Velocity Verlet algorithm. The
        necessary equations of motion for the velocity is
            v(t_i+1) = v(t_i) + /delta t * 0.5(a(t_i+1) + a(t_i))'''
        return self.veloc + 0.5 * self.dt * (self.accel + next_accel)



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

'''
Restricted coupled perturbed Hartree-Fock solver
Modified from pyscf.scf.cphf
'''


import numpy as np
from pyscf import lib
from pyscf.lib import logger


def solve(fvind, mo_energy, mo_occ, h1, s1=None,
          max_cycle=20, tol=1e-9, hermi=False, verbose=logger.WARN):
    '''
    Args:
        fvind : function
            Given density matrix, compute (ij|kl)D_{lk}*2 - (ij|kl)D_{jk}

    Kwargs:
        hermi : boolean
            Whether the matrix defined by fvind is Hermitian or not.
    '''
    if s1 is None:
        return solve_nos1(fvind, mo_energy, mo_occ, h1,
                          max_cycle, tol, hermi, verbose)
    else:
        return solve_withs1(fvind, mo_energy, mo_occ, h1, s1,
                            max_cycle, tol, hermi, verbose)
kernel = solve

# h1 shape is (:,nvir,nocc)
def solve_nos1(fvind, mo_energy, mo_occ, h1,
               max_cycle=20, tol=1e-9, hermi=False, verbose=logger.WARN):
    '''For field independent basis. First order overlap matrix is zero'''
    log = logger.new_logger(verbose=verbose)
    t0 = (logger.process_clock(), logger.perf_counter())

    nkpt = len(h1)
    moloc = np.zeros([nkpt+1], dtype=int)
    for k in range(nkpt):
        moloc[k+1] = moloc[k] + h1[k].size

    occidx = []
    viridx = []
    for k in range(nkpt):
        occidx.append(mo_occ[k] > 0)
        viridx.append(mo_occ[k] == 0)

    e_a = [mo_energy[k][viridx[k]] for k in range(nkpt)]
    e_i = [mo_energy[k][occidx[k]] for k in range(nkpt)]
    e_ai = [1 / lib.direct_sum('a-i->ai', e_a[k], e_i[k]) for k in range(nkpt)]
    mo1base = []
    for k in range(nkpt):
        mo1base.append((h1[k] * -e_ai[k]).ravel())
    mo1base = np.hstack(mo1base)

    def vind_vo(mo1):
        mo1 = mo1.flatten()
        tmp = []
        for k in range(nkpt):
            tmp.append(mo1[moloc[k]:moloc[k+1]].reshape(h1[k].shape))
        v = fvind(tmp)
        for k in range(nkpt):
            v[k] *= e_ai[k]
            v[k] = v[k].ravel()
        return np.hstack(v)

    _mo1 = lib.krylov(vind_vo, mo1base,
                      tol=tol, max_cycle=max_cycle, hermi=hermi, verbose=log).flatten()
    log.timer('krylov solver in CPHF', *t0)
    mo1 = []
    for k in range(nkpt):
        mo1.append(_mo1[moloc[k]:moloc[k+1]].reshape(h1[k].shape))
    return mo1, None

# h1 shape is (:,nocc+nvir,nocc)
def solve_withs1(fvind, mo_energy, mo_occ, h1, s1,
                 max_cycle=20, tol=1e-9, hermi=False, verbose=logger.WARN):
    '''For field dependent basis. First order overlap matrix is non-zero.
    The first order orbitals are set to
    C^1_{ij} = -1/2 S1
    e1 = h1 - s1*e0 + (e0_j-e0_i)*c1 + vhf[c1]

    Kwargs:
        hermi : boolean
            Whether the matrix defined by fvind is Hermitian or not.

    Returns:
        First order orbital coefficients (in MO basis) and first order orbital
        energy matrix
    '''
    log = logger.new_logger(verbose=verbose)
    t0 = (logger.process_clock(), logger.perf_counter())

    nkpt = len(h1)
    ncomp = h1[0].shape[0]
    occidx = []
    viridx = []
    for k in range(nkpt):
        occidx.append(mo_occ[k] > 0)
        viridx.append(mo_occ[k] == 0)

    e_a = [mo_energy[k][viridx[k]] for k in range(nkpt)]
    e_i = [mo_energy[k][occidx[k]] for k in range(nkpt)]
    e_ai = [1 / lib.direct_sum('a-i->ai', e_a[k], e_i[k]) for k in range(nkpt)]
    nocc = np.zeros([nkpt], dtype=int)
    nvir = np.zeros([nkpt], dtype=int)
    nmo  = np.zeros([nkpt], dtype=int)
    moloc = np.zeros([nkpt+1], dtype=int)
    for k in range(nkpt):
        nvir_k, nocc_k = e_ai[k].shape
        nmo_k = nvir_k + nocc_k
        nvir[k] = nvir_k
        nocc[k] = nocc_k
        nmo[k]  = nmo_k
        moloc[k+1] = moloc[k] + nmo_k * nocc_k * ncomp

    mo1base = []
    _mo1base = []
    mo_e1 = []
    for k in range(nkpt):
        mo1base.append(h1[k] - s1[k] * e_i[k])
        mo_e1.append(mo1base[k][:,occidx[k],:].copy())
        mo1base[k][:,viridx[k]] *= -e_ai[k]
        mo1base[k][:,occidx[k]] = -s1[k][:,occidx[k]] * .5
        _mo1base.append(mo1base[k].ravel())
    _mo1base = np.hstack(_mo1base)

    def vind_vo(mo1):
        mo1 = mo1.ravel()
        tmp = []
        for k in range(nkpt):
            tmp.append(mo1[moloc[k]:moloc[k+1]].reshape(-1,nmo[k],nocc[k]))
        v = fvind(tmp)
        for k in range(nkpt):
            v[k][:,viridx[k],:] *= e_ai[k]
            v[k][:,occidx[k],:] = 0
            v[k] = v[k].ravel()
        return np.hstack(v)
    _mo1 = lib.krylov(vind_vo, _mo1base,
                      tol=tol, max_cycle=max_cycle, hermi=hermi, verbose=log)
    mo1 = []
    for k in range(nkpt):
        mo1.append(_mo1[moloc[k]:moloc[k+1]].reshape(-1,nmo[k],nocc[k]))
    log.timer('krylov solver in CPHF', *t0)

    v1mo = fvind(mo1)
    for k in range(nkpt):
        mo1[k][:,viridx[k]] = mo1base[k][:,viridx[k]] - \
                             v1mo[k][:,viridx[k]]*e_ai[k]

    # mo_e1 has the same symmetry as the first order Fock matrix (hermitian or
    # anti-hermitian). mo_e1 = v1mo + u1*lib.direct_sum('i-j->ij',e_i,e_i)
    for k in range(nkpt):
        mo_e1[k] += mo1[k][:,occidx[k]] * lib.direct_sum('i-j->ij', e_i[k], e_i[k])
        mo_e1[k] += v1mo[k][:,occidx[k]]

    return mo1, mo_e1



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

import ctypes
import numpy as np
from pyscf import lib
from pyscf.gto import moleintor
from pyscf.gto.eval_gto import _get_intor_and_comp, BLKSIZE
from pyscf.pbc.gto import _pbcintor
from pyscf.gto.mole import extract_pgto_params, ANG_OF
from pyscf import __config__

EXTRA_PREC = getattr(__config__, 'pbc_gto_eval_gto_extra_precision', 1e-2)

libpbc = _pbcintor.libpbc

def eval_gto(cell, eval_name, coords, comp=None, kpts=None, kpt=None,
             shls_slice=None, non0tab=None, ao_loc=None, cutoff=None,
             out=None, Ls=None, rcut=None):
    r'''Evaluate PBC-AO function value on the given grids,

    Args:
        eval_name : str::

            ==========================  =======================
            Function                    Expression
            ==========================  =======================
            "GTOval_sph"                \sum_T exp(ik*T) |AO>
            "GTOval_ip_sph"             nabla \sum_T exp(ik*T) |AO>
            "GTOval_cart"               \sum_T exp(ik*T) |AO>
            "GTOval_ip_cart"            nabla \sum_T exp(ik*T) |AO>
            ==========================  =======================

        atm : int32 ndarray
            libcint integral function argument
        bas : int32 ndarray
            libcint integral function argument
        env : float64 ndarray
            libcint integral function argument

        coords : 2D array, shape (N,3)
            The coordinates of the grids.

    Kwargs:
        shls_slice : 2-element list
            (shl_start, shl_end).
            If given, only part of AOs (shl_start <= shell_id < shl_end) are
            evaluated.  By default, all shells defined in cell will be evaluated.
        non0tab : 2D bool array
            mask array to indicate whether the AO values are zero.  The mask
            array can be obtained by calling :func:`dft.gen_grid.make_mask`
        cutoff : float
            AO values smaller than cutoff will be set to zero. The default
            cutoff threshold is ~1e-22 (defined in gto/grid_ao_drv.h)
        out : ndarray
            If provided, results are written into this array.

    Returns:
        A list of 2D (or 3D) arrays to hold the AO values on grids.  Each
        element of the list corresponds to a k-point and it has the shape
        (N,nao) Or shape (\*,N,nao).

    Examples:

    >>> cell = pbc.gto.M(a=numpy.eye(3)*4, atom='He 1 1 1', basis='6-31g')
    >>> coords = cell.get_uniform_grids([10,10,10])
    >>> kpts = cell.make_kpts([3,3,3])
    >>> ao_value = cell.pbc_eval_gto("GTOval_sph", coords, kpts)
    >>> len(ao_value)
    27
    >>> ao_value[0].shape
    (1000, 2)
    >>> ao_value = cell.pbc_eval_gto("GTOval_ig_sph", coords, kpts, comp=3)
    >>> print(ao_value.shape)
    >>> len(ao_value)
    27
    >>> ao_value[0].shape
    (3, 1000, 2)
    '''
    if eval_name[:3] == 'PBC':  # PBCGTOval_xxx
        eval_name, comp = _get_intor_and_comp(cell, eval_name[3:], comp)
    else:
        eval_name, comp = _get_intor_and_comp(cell, eval_name, comp)
    eval_name = 'PBC' + eval_name

    atm = np.asarray(cell._atm, dtype=np.int32, order='C')
    bas = np.asarray(cell._bas, dtype=np.int32, order='C')
    env = np.asarray(cell._env, dtype=np.double, order='C')
    natm = atm.shape[0]
    nbas = bas.shape[0]
    if kpts is None:
        if kpt is not None:
            kpts_lst = np.reshape(kpt, (1,3))
        else:
            kpts_lst = np.zeros((1,3))
    else:
        kpts_lst = np.reshape(kpts, (-1,3))
    nkpts = len(kpts_lst)
    ngrids = len(coords)

    if non0tab is None:
        non0tab = np.empty(((ngrids+BLKSIZE-1)//BLKSIZE, nbas),
                              dtype=np.uint8)
# non0tab stores the number of images to be summed in real space.
# Initializing it to 255 means all images should be included
        non0tab[:] = 0xff

    if ao_loc is None:
        ao_loc = moleintor.make_loc(bas, eval_name)
    if shls_slice is None:
        shls_slice = (0, nbas)
    sh0, sh1 = shls_slice
    nao = ao_loc[sh1] - ao_loc[sh0]

    out = np.empty((nkpts,comp,nao,ngrids), dtype=np.complex128)
    coords = np.asarray(coords, order='F')

    if rcut is None:
        deriv = eval_name.count('ip')
        rcut = _estimate_rcut(cell, deriv)
    if Ls is None:
        Ls = get_lattice_Ls(cell, rcut=rcut.max())
        Ls = Ls[np.argsort(lib.norm(Ls, axis=1), kind='stable')]
    expLk = np.exp(1j * np.asarray(np.dot(Ls, kpts_lst.T), order='C'))

    with cell.with_integral_screen(cutoff):
        drv = getattr(libpbc, eval_name)
        drv(ctypes.c_int(ngrids),
            (ctypes.c_int*2)(*shls_slice), ao_loc.ctypes.data_as(ctypes.c_void_p),
            Ls.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(len(Ls)),
            expLk.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(nkpts),
            out.ctypes.data_as(ctypes.c_void_p),
            coords.ctypes.data_as(ctypes.c_void_p),
            rcut.ctypes.data_as(ctypes.c_void_p),
            non0tab.ctypes.data_as(ctypes.c_void_p),
            atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(natm),
            bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(nbas),
            env.ctypes.data_as(ctypes.c_void_p))

    ao_kpts = []
    for k, kpt in enumerate(kpts_lst):
        v = out[k]
        if abs(kpt).sum() < 1e-9:
            v = np.asarray(v.real, order='C')
        v = v.transpose(0,2,1)
        if comp == 1:
            v = v[0]
        ao_kpts.append(v)

    if kpts is None or np.shape(kpts) == (3,):  # A single k-point
        ao_kpts = ao_kpts[0]
    return ao_kpts

pbc_eval_gto = eval_gto

def _estimate_rcut(cell, deriv=0):
    '''Cutoff radius, above which each shell decays to a value less than the
    required precision'''
    es, cs = extract_pgto_params(cell, 'diffused')
    ls = cell._bas[:,ANG_OF]

    vol = cell.vol
    weight_penalty = vol # ~ V[r] * (vol/ngrids) * ngrids
    rad = vol**(-1./3) * cell.rcut + 1
    surface = 4*np.pi * rad**2
    lattice_sum_factor = surface
    precision = cell.precision / max(weight_penalty*lattice_sum_factor, 1)

    norm_ang = ((2*ls+1)/(4*np.pi))**.5
    fac = 2*np.pi/vol * cs*norm_ang/es / precision

    r = cell.rcut
    r = (np.log(fac * r**(ls+1)*(2*es*r)**deriv + 1.) / es)**.5
    r = (np.log(fac * r**(ls+1)*(2*es*r)**deriv + 1.) / es)**.5
    return r

def get_lattice_Ls(cell, nimgs=None, rcut=None, dimension=None, discard=True):
    '''Get lattice-sum vectors for eval_gto
    '''
    if dimension is None:
        # For atoms near the boundary of the cell, it is necessary (even in low-
        # dimensional systems) to include lattice translations in all 3 dimensions.
        if cell.dimension < 2 or cell.low_dim_ft_type == 'inf_vacuum':
            dimension = cell.dimension
        else:
            dimension = 3
    if rcut is None:
        rcut = cell.rcut

    if dimension == 0 or rcut <= 0:
        return np.zeros((1, 3))

    a = cell.lattice_vectors()
    atom_coords = cell.atom_coords()
    scaled_atom_coords = np.linalg.solve(a.T, atom_coords.T).T
    atom_boundary_max = scaled_atom_coords[:,:dimension].max(axis=0)
    atom_boundary_min = scaled_atom_coords[:,:dimension].min(axis=0)
    atom_boundary_max[atom_boundary_max > 1] = 1
    atom_boundary_min[atom_boundary_min <-1] = -1
    atom_bound1 = np.diag(atom_boundary_max).dot(a[:dimension])
    atom_bound2 = np.diag(atom_boundary_min).dot(a[:dimension])

    def find_boundary(a):
        aR = np.vstack([a, atom_bound1, atom_bound2])
        r = np.linalg.qr(aR.T)[1]
        ub = (rcut + abs(r[2,3:]).max()) / abs(r[2,2])
        return ub

    xb = find_boundary(a[[1,2,0]])
    if dimension > 1:
        yb = find_boundary(a[[2,0,1]])
    else:
        yb = 0
    if dimension > 2:
        zb = find_boundary(a)
    else:
        zb = 0
    bounds = np.ceil([xb, yb, zb]).astype(int)
    Ts = lib.cartesian_prod((np.arange(-bounds[0], bounds[0]+1),
                             np.arange(-bounds[1], bounds[1]+1),
                             np.arange(-bounds[2], bounds[2]+1)))
    Ls = np.dot(Ts[:,:dimension], a[:dimension])

    # grids with wrap_around: grids_edge ~ [-.5, .5]
    # regular grids: grids_edge ~ [0, 1]
    grids_edge = lib.cartesian_prod([[-.5, 1.]] * dimension).dot(a[:dimension])
    edge_lb = grids_edge.min(axis=0)
    edge_ub = grids_edge.max(axis=0)

    grids2atm = Ls + atom_coords[:,None,:]
    edge_filter1 = grids2atm > edge_lb
    edge_filter2 = grids2atm < edge_ub
    grids2atm[~edge_filter1[:,:,0],0] -= edge_lb[0]
    grids2atm[~edge_filter1[:,:,1],1] -= edge_lb[1]
    grids2atm[~edge_filter1[:,:,2],2] -= edge_lb[2]
    grids2atm[~edge_filter2[:,:,0],0] -= edge_ub[0]
    grids2atm[~edge_filter2[:,:,1],1] -= edge_ub[1]
    grids2atm[~edge_filter2[:,:,2],2] -= edge_ub[2]
    grids2atm[edge_filter1 & edge_filter2] = 0.
    Ls_mask = (np.linalg.norm(grids2atm[:,:,:dimension], axis=2) < rcut).any(axis=0)
    Ls = Ls[Ls_mask]
    return np.asarray(Ls, order='C')
